"""CompareView: E3SM CompareView application."""

# TODO(decouple): Revisit "quickview" naming/usage in Tauri configs after refactor.
__version__ = "1.2.1"
__author__ = "Kitware Inc."
__license__ = "Apache-2.0"
